#include<limits.h>
#include<stdlib.h>
#include<stdio.h>
#include<malloc.h>
#include<string.h>
#include<unistd.h>
#include<dirent.h>
#include<time.h>
#include<sys/stat.h>

#include "odf.h"

const char bin2base64[]={
	"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	"abcdefghijklmnopqrstuvwxyz"
	"0123456789+/"};

long int rnd_value;

int my_rnd(void){
	long int x,c;
	for(c=0;c<8;c++){
	x=rnd_value&1;
	rnd_value>>=1;
	if(x) rnd_value^=0b10000000001000000000000000000011;
	}
	return rnd_value&63;
}

#ifdef WIN32
char *gettempdir(void){
	unsigned char *cp;
	cp=(unsigned char*)getenv("tmp");
	if(cp) return cp;
	cp=(unsigned char*)getenv("temp");
	return cp;
}

#else
char *gettempdir(void){
	char *cp;
	cp=getenv("TMPDIR");
	if(cp) return cp;
	cp=P_tmpdir;
	if(cp) return cp;
	return "/tmp";
}
#endif

void myMkDir(char *root, char* dir){
	char st[PATH_MAX];
	sprintf(st,"%s/%s",root,dir);
#ifdef WIN32
	if(mkdir(st)) printf("Fehler beim Erstellen von %s\n",st);
#else
	if(mkdir(st,0777)) printf("Fehler beim Erstellen von %s\n",st);
	//else printf("Verzeichnis %s erfolgreich erstellt\n",st);
#endif
}

void delTree(char *root){
	char st[PATH_MAX];
	DIR *d;
	struct dirent *de;
	struct stat sb;
	//printf("leere Verzeichnis %s\n",root);
	d=opendir(root);
	if(!d) return;
	while(de=readdir(d)){
		if(de->d_name[0]!='.'){
			sprintf(st,"%s/%s",root,de->d_name);
			stat(st,&sb);
			if(sb.st_mode&S_IFDIR){
			//printf("Verzeichnis %s\n",st);
			delTree(st);
			}
			else{
			//printf("Datei %s\n",st);
			remove(st);
			}
		}
	}
	closedir(d);
#ifdef WIN32
	rmdir(root);
#else
	remove(root);
#endif
}

int mkTmpDir(char *nam){
	char tmpDir[PATH_MAX];
	char suffix[PATH_MAX];
	time_t tt;
	int versuch=0,r,a,b,c,d;
	tt=time(0);
	rnd_value=tt;
	strcpy(tmpDir,gettempdir());
	do{
		for(c=0;c<20;c++){
			do{
				r=my_rnd();
			} while(r>62);
			suffix[c]=bin2base64[r];
		}
		suffix[c+1]=0;
		sprintf(nam,"%s/%s",tmpDir,suffix);
		//printf("Versuch %d versuche %s anzulegen.\n",++versuch,nam);
#ifdef WIN32
		if(!mkdir(nam)) return(1);	/* erfolgreich */
#else
		if(!mkdir(nam,0777)) return(1);	/* erfolgreich */
#endif
	} while((versuch)<10);
	return 0;	/* fehlgeschlagen */
}

void odf_close(ODF_FILE *odf){
	char st[PATH_MAX];

	if(odf->content){
		fprintf(odf->content,"</office:body></office:document-content>");
		fclose(odf->content);
	}

	if(odf->meta){
		fprintf(odf->meta,"</office:meta></office:document-meta>");
		fclose(odf->meta);
	}

	if(odf->styles){
	fprintf(odf->styles,
	"<office:styles>"
	"<number:number-style style:name=\"N0\">"
	"<number:number number:min-integer-digits=\"1\"/>"
	"</number:number-style>"
	"<style:style style:name=\"Default\" style:family=\"table-cell\" style:data-style-name=\"N0\">"
	"<style:table-cell-properties style:vertical-align=\"automatic\" fo:background-color=\"transparent\"/>"
	"<style:text-properties fo:color=\"#000000\" style:font-name=\"Arial\" style:font-name-asian=\"Arial\" style:font-name-complex=\"Arial\" fo:font-size=\"10pt\" style:font-size-asian=\"10pt\" style:font-size-complex=\"10pt\"/>"
	"</style:style>"
	"</office:styles>"

	"<office:automatic-styles>"
	"</office:automatic-styles>"

	"<office:master-styles>"
	"<style:master-page style:name=\"mp1\" style:page-layout-name=\"pm1\">"
	"<style:header/>"
	"<style:header-left style:display=\"false\"/>"
	"<style:footer/>"
	"<style:footer-left style:display=\"false\"/>"
	"</style:master-page>"
	"</office:master-styles>"

	"</office:document-styles>");
		fclose(odf->styles);
	}
	//printf("schließe Datei %s\n",odf->name);
	chdir(odf->root);
	sprintf(st,"zip -r %s .",odf->name);
	system(st);
	chdir(odf->old_cwd);
	delTree(odf->root);
	free(odf);
}

ODF_FILE *odf_open(char *name, int type){
	ODF_FILE *odf;
	FILE *dst;
	char st[PATH_MAX];
	time_t tt;
	struct tm *t;
	odf=(ODF_FILE*)calloc(1,sizeof(ODF_FILE));
	if(!odf) return 0;
	getcwd(odf->old_cwd,PATH_MAX);
	if((name[0]=='/')||(name[1]==':')){	/* absoluter Pfadname */
		strcpy(odf->name,name);
	}
	else{	/* relativer Pfadname: cwd vorsetzen */
		getcwd(odf->name,PATH_MAX);
		strcat(odf->name,"/");
		strcat(odf->name,name);
	}
	remove(odf->name);
	odf->type=type;
	mkTmpDir(odf->root);
	myMkDir(odf->root,"META-INF");
	sprintf(st,"%s/mimetype",odf->root);
	dst=fopen(st,"w");
	fprintf(dst,"%s",mimetypes[odf->type]);
	fclose(dst);

	sprintf(st,"%s/META-INF/manifest.xml",odf	->root);
	dst=fopen(st,"w");
	fprintf(dst,
	"<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n"
	"<manifest:manifest xmlns:manifest=\"urn:oasis:names:tc:opendocument:xmlns:manifest:1.0\">"
	"<manifest:file-entry manifest:full-path=\"/\" manifest:media-type=\"%s\"/>"
	"<manifest:file-entry manifest:full-path=\"styles.xml\" manifest:media-type=\"text/xml\"/>"
	"<manifest:file-entry manifest:full-path=\"content.xml\" manifest:media-type=\"text/xml\"/>"
	"<manifest:file-entry manifest:full-path=\"mimetype\" manifest:media-type=\"text/plain\"/>"
	"<manifest:file-entry manifest:full-path=\"META-INF/manifest.xml\" manifest:media-type=\"text/xml\"/>"
	"</manifest:manifest>",mimetypes[odf->type]);
	fclose(dst);

	/* meta.xml */
	tt=time(0);
	t=gmtime(&tt);
	sprintf(st,"%s/meta.xml",odf->root);
	odf->meta=fopen(st,"w");
	sprintf(st,"%04d-%02d-%02dT%02d:%02d:%02dZ",t->tm_year+1900,t->tm_mon+1,t->tm_mday,t->tm_hour,t->tm_min,t->tm_sec);
	fprintf(odf->meta,"<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n"
	"<office:document-meta xmlns:office=\"urn:oasis:names:tc:opendocument:xmlns:office:1.0\" xmlns:meta=\"urn:oasis:names:tc:opendocument:xmlns:meta:1.0\" xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" office:version=\"1.1\">"
	"<office:meta>"
		"<meta:creation-date>%s</meta:creation-date>"
		"<dc:date>%s</dc:date>"
		"<meta:editing-duration>PT0S</meta:editing-duration>",st,st);

	/* styles.xml */
	sprintf(st,"%s/styles.xml",odf->root);
	odf->styles=fopen(st,"w");
	fprintf(odf->styles,
	"<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n"
	"<office:document-styles xmlns:table=\"urn:oasis:names:tc:opendocument:xmlns:table:1.0\" xmlns:office=\"urn:oasis:names:tc:opendocument:xmlns:office:1.0\" xmlns:text=\"urn:oasis:names:tc:opendocument:xmlns:text:1.0\" xmlns:style=\"urn:oasis:names:tc:opendocument:xmlns:style:1.0\" xmlns:draw=\"urn:oasis:names:tc:opendocument:xmlns:drawing:1.0\" xmlns:fo=\"urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:number=\"urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0\" xmlns:svg=\"urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0\">"
	"<office:font-face-decls>"
	);


	/* content.xml */
	sprintf(st,"%s/content.xml",odf->root);
	odf->content=fopen(st,"w");
	fprintf(odf->content,
	"<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n"
	"<office:document-content xmlns:table=\"urn:oasis:names:tc:opendocument:xmlns:table:1.0\" xmlns:office=\"urn:oasis:names:tc:opendocument:xmlns:office:1.0\" xmlns:text=\"urn:oasis:names:tc:opendocument:xmlns:text:1.0\" xmlns:style=\"urn:oasis:names:tc:opendocument:xmlns:style:1.0\" xmlns:draw=\"urn:oasis:names:tc:opendocument:xmlns:drawing:1.0\" xmlns:fo=\"urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:number=\"urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0\" xmlns:svg=\"urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0\">"
	"<office:font-face-decls>"
	);
	return odf;
}

void odf_addMeta(ODF_FILE *odf, char *meta, char *value){
	fprintf(odf->meta,"<%s>%s</%s>",meta,value,meta);
}

void odf_addFont(ODF_FILE *odf,char *name){
	char st[2048];
	sprintf(st,"<style:font-face style:name=\"%s\" svg:font-family=\"%s\"/>",name,name);
	fprintf(odf->content,"%s",st);
	fprintf(odf->styles,"%s",st);
}

void odf_endFonts(ODF_FILE *odf){
	fprintf(odf->content,"</office:font-face-decls><office:automatic-styles>");
	fprintf(odf->styles,"</office:font-face-decls>");
}

void odf_endAutoStyles(ODF_FILE *odf){
	fprintf(odf->content,"</office:automatic-styles><office:body>");
}

void odf_writeString(ODF_FILE *odt, char *Line){
	while(*Line){
		switch(*Line){
			case '\t':
				fprintf(odt->content,"<text:tab/>");
			break;

			case '\n':
				fprintf(odt->content,"<text:line-break/>");
			break;

			case '<':
				fprintf(odt->content,"&lt;");
			break;

			case '>':
				fprintf(odt->content,"&gt;");
			break;

			case '&':
				fprintf(odt->content,"&amp;");
			break;

			default:
				fprintf(odt->content,"%c",*Line);
			break;
		}
		Line++;
	}
}

void outTextStyles(ODF_FILE *odf, int flags){
	if(flags&TS_ITALIC) fprintf(odf->content," fo:font-style=\"italic\"");
	if(flags&TS_BOLD) fprintf(odf->content," fo:font-weight=\"bold\"");
	if(flags&TS_UNDERLINE) fprintf(odf->content," style:text-underline-style=\"solid\" style:text-underline-width=\"auto\" style:text-underline-color=\"font-color\"");
}


/* Auto-Formate */

/* Nummenformate */

void odf_addPercentageFormat(ODF_FILE *odf,int nr,int decimal,int digits){
	fprintf(odf->content,"<number:percentage-style style:name=\"N%d\">"
	"<number:number number:decimal-places=\"%d\""
	" number:min-integer-digits=\"%d\"/>"
	"<number:text>%%</number:text></number:percentage-style>",nr,decimal,digits);
}

void odf_addNumberFormat(ODF_FILE *odf,int nr,int decimal,int digits,char *unit){
	fprintf(odf->content,
	"<number:number-style style:name=\"N%d\">"
	"<number:number number:decimal-places=\"%d\" "
	"number:min-integer-digits=\"%d\"/>",nr,decimal,digits);
	if(unit)fprintf(odf->content,"<number:text>%s</number:text>",unit);
	fprintf(odf->content,"</number:number-style>");
}

/* Absatzformat */

void odf_addParagraphStyle(ODF_FILE *odt, int nr, int fontSize, int flags, int line_height){
	int horz;
	horz=flags&3;
	fprintf(odt->content,"<style:style style:name=\"P%d\" style:family=\"paragraph\" style:parent-style-name=\"Standard\">",nr);

	fprintf(odt->content,"<style:paragraph-properties");
	if(line_height!=100) fprintf(odt->content," fo:line-height=\"%d%%\"",line_height);

	switch(horz){
		case CF_LEFT:
		fprintf(odt->content," fo:text-align=\"left\"/>");
		break;
		case CF_CENTER:
		fprintf(odt->content," fo:text-align=\"center\"/>");
		break;
		case CF_RIGHT:
		fprintf(odt->content," fo:text-align=\"right\"/>");
		break;
		default:
		fprintf(odt->content," fo:text-align=\"left\"/>");
		break;
	}

	fprintf(odt->content,"<style:text-properties");
	outTextStyles(odt,flags);
	fprintf(odt->content," fo:font-size=\"%dpt\"/>",fontSize);
	fprintf(odt->content,"</style:style>");

}

void odf_addRowFormat(ODF_FILE *odt,int format, double height){
	fprintf(odt->content,
	"<style:style style:name=\"ro%d\" style:family=\"table-row\">"
	"<style:table-row-properties style:row-height=\"%5.3lfcm\"/>"
	"</style:style>",format,height);
}

void odf_addColFormat(ODF_FILE *odf, int format, double width){
	fprintf(odf->content,
	"<style:style style:name=\"co%d\" style:family=\"table-column\">"
	"<style:table-column-properties style:column-width=\"%6.3lfin\"/>"
	"</style:style>",format,width/2.54);
}

void resetCellstyle(void){
	memset(&cellStyle,0,sizeof(cellStyle));
	cellStyle.text_size=12;
}

void setBorder(int mask, int style, double width, int color){
	if(mask&BM_TOP){
		cellStyle.top.style=style;
		cellStyle.top.width=width;
		cellStyle.top.color=color;
	}
	if(mask&BM_BOTTOM){
		cellStyle.bottom.style=style;
		cellStyle.bottom.width=width;
		cellStyle.bottom.color=color;
	}
	if(mask&BM_LEFT){
		cellStyle.left.style=style;
		cellStyle.left.width=width;
		cellStyle.left.color=color;
	}
	if(mask&BM_RIGHT){
		cellStyle.right.style=style;
		cellStyle.right.width=width;
		cellStyle.right.color=color;
	}
}

void out_borderStyle(ODF_FILE *odf, border *brd, char *name){
	fprintf(odf->content," fo:border-%s=\"",name);
	if(brd->style){
	fprintf(odf->content,"%4.2lfpt ",brd->width);
	}
	fprintf(odf->content,"%s",borderstylenam[brd->style]);
	if(brd->style){
	fprintf(odf->content," #%06x",brd->color);
	}
	fprintf(odf->content,"\"");
}

void add_cellStyle(ODF_FILE *odf,int format){
	fprintf(odf->content,"<style:style style:name=\"ce%d\" style:family=\"table-cell\"",format);
	if(cellStyle.data_style) fprintf(odf->content," style:data-style-name=\"N%d\"",cellStyle.data_style);
	fprintf(odf->content,"><style:table-cell-properties");
	switch(cellStyle.vert_align){
		case VA_TOP:
		fprintf(odf->content," style:vertical-align=\"top\"");
		break;
		case VA_MIDDLE:
		fprintf(odf->content," style:vertical-align=\"middle\"");
		break;
		case VA_BOTTOM:
		fprintf(odf->content," style:vertical-align=\"bottom\"");
		break;
	}
	out_borderStyle(odf,&cellStyle.left,"left");
	out_borderStyle(odf,&cellStyle.right,"right");
	out_borderStyle(odf,&cellStyle.top,"top");
	out_borderStyle(odf,&cellStyle.bottom,"bottom");
	if(cellStyle.bg_color) fprintf(odf->content," fo:background-color=\"#%06x\"",cellStyle.bg_color);

	fprintf(odf->content,"/>");
	fprintf(odf->content,"<style:paragraph-properties");
	fprintf(odf->content," fo:text-align=\"%s\"",h_alig_names[cellStyle.horz_align]);
	if((cellStyle.horz_align==HO_LEFT)||(cellStyle.horz_align==HO_JUSTIFY))
		fprintf(odf->content," fo:margin-left=\"%dmm\"",cellStyle.h_margin);
	if((cellStyle.horz_align==HO_RIGHT)||(cellStyle.horz_align==HO_JUSTIFY))
		fprintf(odf->content," fo:margin-right=\"%dmm\"/>",cellStyle.h_margin);
	fprintf(odf->content,"/>");
	fprintf(odf->content,"<style:text-properties");
	fprintf(odf->content," fo:color=\"#%06x\"",cellStyle.text_color);
	if(strlen(cellStyle.font_name)) fprintf(odf->content," style:font-name=\"%s\"",cellStyle.font_name);
	fprintf(odf->content," fo:font-size=\"%dpt\"",cellStyle.text_size);
	outTextStyles(odf,cellStyle.text_styles);
	fprintf(odf->content,"/>");
	fprintf(odf->content,"</style:style>");
}


void odf_addTableFormat(ODF_FILE *odf, int format, double width){
	fprintf(odf->content,
	"<style:style style:name=\"ta%d\" style:family=\"table\">"
	"<style:table-properties style:width=\"%4.2lfcm\" fo:margin-left=\"0cm\" table:align=\"center\"/>"
	"</style:style>",format,width);
}

/* Objekt: Text */

void odf_writeParagraph(ODF_FILE *odt, char *Line, int paragraphStyle){
	fprintf(odt->content,"<text:p");
	if(paragraphStyle) fprintf(odt->content," text:style-name=\"P%d\"",paragraphStyle);
	fprintf(odt->content,">");
	odf_writeString(odt,Line);
	fprintf(odt->content,"</text:p>");
}

/* Objekt: Tabelle */

void odf_writeCell(ODF_FILE *odf,char *string,int cellstyle,int paragraphStyle){
	fprintf(odf->content,"<table:table-cell");
	if(cellstyle) fprintf(odf->content," table:style-name=\"ce%d\"",cellstyle);
	fprintf(odf->content," office:value-type=\"string\">");
	odf_writeParagraph(odf,string,paragraphStyle);
	fprintf(odf->content,"</table:table-cell>"
	);
}

int odf_writeDoubleCell(ODF_FILE *base,double value, int cellstyle, int precision)
{
	char st1[100],st2[100],*cp;
	sprintf(st1,"%.*lf",precision,value);
	strcpy(st2,st1);
//	if(cp=strchr(st2,'.')) *cp=',';
	fprintf(base->content,"<table:table-cell");
	if(cellstyle) fprintf(base->content," table:style-name=\"ce%d\"",cellstyle);
	fprintf(base->content," office:value=\"%s\" office:value-type=\"float\"><text:p>%s</text:p></table:table-cell>",st1,st2);
	return 0;
}

int odf_writeIntCell(ODF_FILE *base,int value, int cellstyle)
{
	fprintf(base->content,"<table:table-cell");
	if(cellstyle) fprintf(base->content," table:style-name=\"ce%d\"",cellstyle);
	fprintf(base->content," office:value=\"%d\" office:value-type=\"float\"><text:p>%d</text:p></table:table-cell>",value,value);
	return 0;
}

int odf_writePercent0Cell(ODF_FILE *base,double value, int cellstyle)
{
	fprintf(base->content,"<table:table-cell");
	if(cellstyle) fprintf(base->content," table:style-name=\"ce%d\"",cellstyle);
	fprintf(base->content," office:value-type=\"percentage\" office:value=\"%4.2f\"><text:p>%3.1lf%%</text:p></table:table-cell>",value,(double)value*100);
	return 0;
}

int odf_writePercent2Cell(ODF_FILE *base,double value, int cellstyle)
{
	fprintf(base->content,"<table:table-cell");
	if(cellstyle) fprintf(base->content," table:style-name=\"ce%d\"",cellstyle);
	fprintf(base->content," office:value-type=\"percentage\" office:value=\"%6.4f\"><text:p>%5.3lf%%</text:p></table:table-cell>",value,(double)value*100);
	return 0;
}

void odf_startTable(ODF_FILE *odf, char *name, int tableStyle){
	fprintf(odf->content,"<table:table table:name=\"%s\"",name);
	if(tableStyle) fprintf(odf->content," table:style-name=\"ta%d\"",tableStyle);
	fprintf(odf->content,">");
}

void odf_endTable(ODF_FILE *odf){
	fprintf(odf->content,"</table:table>");
}

void odf_startColumns(ODF_FILE *odf){
	fprintf(odf->content,"<table:table-columns>");
}

void odf_endColumns(ODF_FILE *odf){
	fprintf(odf->content,"</table:table-columns>");
}

void odf_Column(ODF_FILE *odf,int colStyle, int cellStyle){
	fprintf(odf->content,"<table:table-column");
	if(colStyle) fprintf(odf->content," table:style-name=\"co%d\"",colStyle);
	if(cellStyle) fprintf(odf->content," table:default-cell-style-name=\"ce%d\"",cellStyle);
	fprintf(odf->content,"/>");
}

int odf_ColumnGroup(ODF_FILE *odf,int colStyle,int count,int cellStyle)
{
	fprintf(odf->content,"<table:table-column-group table:display=\"false\">");
	fprintf(odf->content,"<table:table-column table:style-name=\"co%d\" table:number-columns-repeated=\"%d\"",colStyle,count);
	if(cellStyle) fprintf(odf->content," table:default-cell-style-name=\"ce%d\"",cellStyle);
	fprintf(odf->content," table:visibility=\"collapse\"/></table:table-column-group>");
	return 0;
}

void odf_startRow(ODF_FILE *odf, int rowFormat){
	fprintf(odf->content,"<table:table-row");
	if(rowFormat) fprintf(odf->content," table:style-name=\"ro%d\"",rowFormat);
	fprintf(odf->content,">");
}

void odf_endRow(ODF_FILE *odf){
	fprintf(odf->content,"</table:table-row>");
}

void odtStartTableHeaderRows(ODF_FILE *odt){
	fprintf(odt->content,
	"<table:table-header-rows>"
	);
}

void odtEndTableHeaderRow(ODF_FILE *odt){
	fprintf(odt->content,
	"</table:table-header-rows>"
	);
}

/* Dokumentenrahmen */

void odt_startDoc(ODF_FILE *odt){
	fprintf(odt->content,"<office:text>");
}

void odt_endDoc(ODF_FILE *odt){
	fprintf(odt->content,"</office:text>");
}

void ods_startSheet(ODF_FILE *odf){
	fprintf(odf->content,"<office:spreadsheet>");
}

void ods_endSheet(ODF_FILE *odf){
	fprintf(odf->content,"</office:spreadsheet>");
}
